import bs
import bsTeamGame
import bsMap
import bsUtils
import bsSpaz
import bsScoreBoard
import weakref
import bsFlag
import bsPowerup

def getGames():
    return [ChosenOneGame]

class ChosenOneGame(bsTeamGame.TeamGameActivity):

    @classmethod
    def getName(c):
        return 'Chosen One'

    @classmethod
    def getDescription(c,sessionType):
        return ('Be the chosen one for a length of time to win.\n'
                'Kill the chosen one to become it.')

    @classmethod
    def getSupportedSessionTypes(self):
        return ['freeForAll','teams']

    @classmethod
    def getSupportedMaps(self,sessionType):
        return bsMap.getMapList("keepAway")

    @classmethod
    def getSettings(self,sessionType):
        return [("Chosen One Time",{'minValue':10,'default':30,'increment':10}),
                ("Chosen One Gets Gloves",{'default':True}),
                ("Chosen One Gets Shield",{'default':False}),
                ("Time Limit",{'choices':[('None',0),('1 Minute',60),
                                        ('2 Minutes',120),('5 Minutes',300),
                                        ('10 Minutes',600),('20 Minutes',1200)],'default':0}),
                ("Respawn Times",{'choices':[('Shorter',0.25),('Short',0.5),('Normal',1.0),('Long',2.0),('Longer',4.0)],'default':1.0}),
                ("Epic Mode",{'default':False})]

    def __init__(self,session,level,settings):
        bsTeamGame.TeamGameActivity.__init__(self,session,level,settings)
        if self.settings['Epic Mode']: self._isSlowMotion = True
        self._scoreBoard = bsScoreBoard.ScoreBoard(self)
        self._chosenOne = None
        self._swipSound = bs.getSound("swip")
        self._countDownSounds = {10:bs.getSound('announceTen'),
                                 9:bs.getSound('announceNine'),
                                 8:bs.getSound('announceEight'),
                                 7:bs.getSound('announceSeven'),
                                 6:bs.getSound('announceSix'),
                                 5:bs.getSound('announceFive'),
                                 4:bs.getSound('announceFour'),
                                 3:bs.getSound('announceThree'),
                                 2:bs.getSound('announceTwo'),
                                 1:bs.getSound('announceOne')}

    def getInstanceDescription(self):
        return 'There can be only one.'

    def onTransitionIn(self):
        bsTeamGame.TeamGameActivity.onTransitionIn(self, music='Epic' if self.settings['Epic Mode'] else 'Chosen One')

    def onTeamJoined(self,team):
        team.gameData['timeRemaining'] = self.settings["Chosen One Time"]
        self._updateScoreBoard()

    def onBegin(self):
        bsTeamGame.TeamGameActivity.onBegin(self,nameSound=bs.getSound('announceChosenOne'))
        self.setupStandardTimeLimit(self.settings['Time Limit'])
        self._flagSpawnPos = self._level.getFlagPosition(None)
        self.projectFlagStand(self._flagSpawnPos)
        self._setChosenOne(None)

        p = self._flagSpawnPos
        self._updateTimer = bs.Timer(1000,call=bs.WeakCall(self._tick),repeat=True)

        m = self._resetRegionMaterial = bs.Material('Reset Region')
        m.addAction(conditions=("dstIsMaterial",bs.playerMaterial),
                    actions=(("partMod","collide",True),
                             ("partMod","physical",False),
                             ("call","atConnect",bs.WeakCall(self._handleResetCollide))))

        self._resetRegion = bs.newNode("region",
                                       owner=self,
                                       attrs={'translate':(p[0],p[1]+0.75,p[2]),
                                              'scale': (0.5,0.5,0.5),
                                              'type': "sphere",
                                              'materials':(m.materialID,)
                                              })

    def _getChosenOne(self):
        return None if self._chosenOne is None else self._chosenOne()

    def _handleResetCollide(self):
        # if we have a chosen one ignore these
        if self._getChosenOne() is not None: return
        try: player = bs.getCollideValues("opposingNode").getDelegate().getPlayer()
        except Exception: return
        self._setChosenOne(player)
        
    def _flashFlagSpawn(self):
        light = bs.newNode("light",
                           owner=self,
                           attrs={'translate':self._flagSpawnPos,'color':(1,1,1),
                                  'radius':0.3,'heightAttenuated':False})
        bsUtils.animate(light,"intensity",{0:0,250:0.5,500:0},loop=True)
        bs.gameTimer(1000,light.delete)

    def _tick(self):

        # give the chosen one points
        player = self._getChosenOne()
        if player is not None:

            # this shouldnt happen but just in case..
            if not player.isAlive():
                self._setChosenOne(None)
            else:
                scoringTeam = player.getTeam()
                self.scoreSet.playerScored(player,3,screenMessage=False,display=False)
                #if scoringTeam.gameData['timeRemaining'] > 0: bs.playSound(self._tickSound)
                scoringTeam.gameData['timeRemaining'] = max(0,scoringTeam.gameData['timeRemaining']-1)

                # show the count over their head
                try:
                    if scoringTeam.gameData['timeRemaining'] > 0:
                        player.actor.setScoreText(str(scoringTeam.gameData['timeRemaining']))
                except Exception: pass
                   
                self._updateScoreBoard()

                # announce numbers we have sounds for
                try: bs.playSound(self._countDownSounds[scoringTeam.gameData['timeRemaining']])
                except Exception: pass

                # winner
                if scoringTeam.gameData['timeRemaining'] <= 0:
                    self.endGame()

    def endGame(self):
        results = bsTeamGame.Results(scoreName='Time Held')
        for team in self.teams.values(): results.setTeamScore(team,self.settings['Chosen One Time'] - team.gameData['timeRemaining'])
        self.end(results=results,announceDelay=0)


    def _setChosenOne(self,player):
        for p in self.players.values(): p.gameData['chosenLight'] = None
        bs.playSound(self._swipSound)
        if player is None:
            self._flag = bsFlag.Flag(self,bs.Actor.UNRETAINED,
                                     color=(1,0.9,0.2),
                                     translate=self._flagSpawnPos,
                                     touchable=False)
            self._chosenOne = None

            l = bs.newNode('light',
                           owner=self._flag.node,
                           attrs={"translate": self._flagSpawnPos,
                                  "intensity":0.6,
                                  "heightAttenuated":False,
                                  "volumeIntensityScale":0.1,
                                  "radius":0.1,
                                  "color": (0.8,0.8,1.2)})

            self._flashFlagSpawn()
        else:
            if player.actor is not None:
                self._flag = None
                self._chosenOne = weakref.ref(player)

                if player.actor.node.exists():
                    if self.settings['Chosen One Gets Shield']: player.actor.handleMessage(bsPowerup.PowerupMessage('shield'))
                    if self.settings['Chosen One Gets Gloves']: player.actor.handleMessage(bsPowerup.PowerupMessage('punch'))
                    l = player.gameData['chosenLight'] = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                                                      'light',{"intensity":0.6,
                                                                               "heightAttenuated":False,
                                                                               "volumeIntensityScale":0.1,
                                                                               "radius":0.13,
                                                                               "color": (0.8,0.8,1.2)})
                    
                    bsUtils.animate(l.node,'intensity',{0:1.0, 250:0.0, 500:1.0},loop=True)
                    player.actor.node.connectAttr('translate',l.node,'translate')


    def handleMessage(self,m):
        if isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            bsTeamGame.TeamGameActivity.handleMessage(self,m) # augment default
            player = m.spaz.getPlayer()
            if player is self._getChosenOne():
                killerPlayer = m.getKillerPlayer()
                self._setChosenOne(None if (killerPlayer is None or killerPlayer is player or not killerPlayer.isAlive()) else killerPlayer)
            self.respawnPlayer(player)
        else: bsTeamGame.TeamGameActivity.handleMessage(self,m)

    def _updateScoreBoard(self):
        for team in self.teams.values():
            self._scoreBoard.setTeamValue(team,team.gameData['timeRemaining'],self.settings['Chosen One Time'],countdown=True)
