import bs
import bsTeamGame
import bsSpaz
import random
import bsBomb

def getGames():
    return [MeteorShowerGame]

class MeteorShowerGame(bsTeamGame.TeamGameActivity):

    @classmethod
    def getName(c):
        return 'Meteor Shower'

    @classmethod
    def getDescription(c,sessionType):
        return 'Dodge the falling bombs.'

    # we're currently hard-coded for one map..
    @classmethod
    def getSupportedMaps(self,sessionType):
        return ['Rampage']

    # we can be played in either free-for-all or teams mode
    # ( free-for-all is really just a special form of teams
    # where each player gets their own team)
    @classmethod
    def getSupportedSessionTypes(self):
        return ['freeForAll','teams']


    # called when our game is transitioning in but not ready to start..
    # ..we can go ahead and set our music and whatnot
    def onTransitionIn(self):
        bsTeamGame.TeamGameActivity.onTransitionIn(self, music='Survival')


    # called when our game actually starts
    def onBegin(self):

        # we dont want normal powerups since most of them won't apply to us
        bsTeamGame.TeamGameActivity.onBegin(self,doStandardPowerups=False)

        # drop a wave every few seconds.. and every so often drop the time between waves
        self._meteorTime = 2000
        self._decrementTimer = bs.Timer(5000,bs.WeakCall(self._decrementMeteorTime),repeat=True)

        # kick off the first wave in a few seconds
        bs.gameTimer(3000,bs.WeakCall(self._setMeteorTimer))

        
    # overriding the default character spawning..
    def spawnPlayer(self,player):

        spaz = self.spawnSpazForPlayer(player)

        # lets reconnect this player's controls to this
        # spaz but *without* the ability to attack or pick stuff up
        spaz.connectToPlayer(player.playerID,
                             enablePunch=False,
                             enableBomb=False,
                             enablePickUp=False)

        # also lets have them make some noise when they die..
        spaz.playBigDeathSound = True

        # lastly let's store the current game-time when this player was born
        # so we can measure how long they lived.
        # 'data' is a special empty dictionary we can use for game-specific player data
        player.gameData['birthTime'] = bs.getGameTime()

        return spaz


    # many high-level game events come through this method
    def handleMessage(self,m):

        if isinstance(m,bsSpaz.PlayerSpazDeathMessage):

            bsTeamGame.TeamGameActivity.handleMessage(self,m) # (augment standard behavior)

            # record the player's moment of death
            player = m.spaz.getPlayer()
            if player is not None: # need to check this; player could have left..
                player.gameData['deathTime'] = bs.getGameTime()
                
            # in one second lets end the game if there are no players left..
            bs.gameTimer(1000,bs.WeakCall(self._checkEndGame))

        else:
            # default handler:
            bsTeamGame.TeamGameActivity.handleMessage(self,m)

    def _checkEndGame(self):
        livingTeamCount = 0
        for team in self.teams.values():
            for player in team.players.values():
                if player.isAlive():
                    livingTeamCount += 1
                    break
        if livingTeamCount <= 1:
            self.endGame()
        
    def _setMeteorTimer(self):
        self._meteorTimer = bs.Timer(int((1.0+0.2*random.random())*self._meteorTime),
                                     bs.WeakCall(self._dropBombCluster))
        
    def _dropBombCluster(self):

        # random note: code like this is a handy way to plot out extents and debug things
        if False:
            bs.newNode('locator',owner=self,attrs={'translate':(8,6,-5.5)})
            bs.newNode('locator',owner=self,attrs={'translate':(8,6,-2.3)})
            bs.newNode('locator',owner=self,attrs={'translate':(-7.3,6,-5.5)})
            bs.newNode('locator',owner=self,attrs={'translate':(-7.3,6,-2.3)})

        # drop several bombs in series..
        delay = 0
        for i in range(random.randrange(1,3)):
            # drop them somewhere within our bounds with velocity pointing toward the opposite side
            pos = (-7.3+15.3*random.random(),11,-5.5+2.1*random.random())
            vel = ((-5.0+random.random()*30.0) * (-1.0 if pos[0] > 0 else 1.0), -4.0,0)
            bs.gameTimer(delay,bs.WeakCall(self._dropBomb,pos,vel))
            delay += 100
        self._setMeteorTimer()

    def _dropBomb(self,position,velocity):
        b = bsBomb.Bomb(self,bs.Actor.RETAINED,
                        translate=position,
                        velocity=velocity)

    def _decrementMeteorTime(self):
        self._meteorTime = max(10,int(self._meteorTime*0.9))
        
    def endGame(self):

        # mark 'death-time' as current for any still-living players
        # and award players points for how long they lasted.
        # (these per-player scores are only meaningful in team-games)
        for team in self.teams.values():
            for player in team.players.values():
                if 'deathTime' not in player.gameData:
                    player.gameData['deathTime'] = bs.getGameTime()+1
                    score = 50  # award survivors extra
                else:
                    score = 0
                score += (player.gameData['deathTime']-player.gameData['birthTime'])/1000
                self.scoreSet.playerScored(player,score,screenMessage=False)

        # ok now calc game results: set a score for each team and then tell the game to end
        results = bsTeamGame.Results(scoreName='Survived',scoreType='seconds')

        # remember that 'free-for-all' mode is simply a special form of 'teams' mode
        # where each player gets their own team, so we can just always deal in teams
        # and have all cases covered
        for team in self.teams.values():

            # set the team score to the max time survived by any player on that team
            longestLife = 0
            for player in team.players.values():
                longestLife = max(longestLife,(player.gameData['deathTime'] - player.gameData['birthTime'])/1000)
            results.setTeamScore(team,longestLife)

        self.end(results=results)
